// hlavickove subory *******************************************************
#include "includes.h"

// konstruktor *************************************************************
CMath3d::CMath3d()
{

}

// destruktor **************************************************************
CMath3d::~CMath3d()
{

}

// kolizia svetelneho luca a trojuholnika **********************************
bool CMath3d::ray_tri_collision(vec3 P,vec3 L,vec3 N,float D,vec4 CN0,vec4 CN1,vec4 CN2)
{
 // vektor P -> L (bod -> poloha svetelneho zdroja)
 v0=L-P; v0.normalize();
 v1=P-L; v1.normalize();

 // skalarny sucin normaly trojuholnika a svetelneho luca
 v2=N;
 a=v2.dot(v0);
 b=v2.dot(v1);

 // v pripade, ak je luc kolmy na normalu trojuholnika, nedoslo ku kolizii
 if (a==0.0f) return(false);

 // projekcia bodu do roviny definovanej trojuholnikom
 a=-(N.x*P.x+N.y*P.y+N.z*P.z+D)/a;
 b=-(N.x*L.x+N.y*L.y+N.z*L.z+D)/b;

 // ak je rovina, ktoru tvori trojuholnik medzi bodom a svetelnym zdrojom,
 // je potrebne otestovat ci priesecnik luca a roviny lezi v trojuholniku
 if ((a>0.0f)&&(b>0.0f)) {
  // priesecnik luca a roviny tvorenej trojuholnikom
  v1=P+v0*a;
  // ak bod lezi v trojuholniku, doslo ku kolizii
  return(point_in_triangle(v1,CN0,CN1,CN2));
 }

 // nedoslo ku kolizii
 return(false);
}

// test polohy bodu a trojuholnika *****************************************
bool CMath3d::point_in_triangle(vec3 P,vec4 CN0,vec4 CN1,vec4 CN2)
{
 // ak bod lezi za stenou trojuholnika => nelezi v trojuholniku
 if ((P.x*CN0.x+P.y*CN0.y+P.z*CN0.z+CN0.w)>0.0f) return(false);
 if ((P.x*CN1.x+P.y*CN1.y+P.z*CN1.z+CN1.w)>0.0f) return(false);
 if ((P.x*CN2.x+P.y*CN2.y+P.z*CN2.z+CN2.w)>0.0f) return(false);

 // bod lezi pred vsetkymi troma stenami => lezi v trojuholniku
 return(true);
}

// kolizia bodu so stranami trojuholnika ***********************************
bool CMath3d::edge_collision(vec3 *P,vec3 *S,vec3 A,vec3 B,vec3 C,float R)
{
 // 1. strana
 v0=point_on_line(A,B,*P); v1=*P-v0;
 if (v1.magnitude()<R) { v1.normalize(); *P=(v0+v1*R); *S=v0; return(true); }
 // 2. strana
 v0=point_on_line(B,C,*P); v1=*P-v0;
 if (v1.magnitude()<R) { v1.normalize(); *P=(v0+v1*R); *S=v0; return(true); }
 // 3. strana
 v0=point_on_line(C,A,*P); v1=*P-v0;
 if (v1.magnitude()<R) { v1.normalize(); *P=(v0+v1*R); *S=v0; return(true); }

 // nedoslo ku kolizii
 return(false);
}

// vypocet polohy bodu na ciare A-B, ktory je najblizsie k bodu C **********
vec3 CMath3d::point_on_line(vec3 A,vec3 B,vec3 C)
{
 v0=C-A; // vektor A->C
 v1=B-A; // vektor A->B
 v2=v1;  // kopia vektora A->B

 v1.normalize(); // normalizacia vektora A->B

 // skalarny sucin vektorov A->C a A->B
 a=v0.dot(v1);

 // bod C lezi pred zaciatkom ciary => najblizsim bodom k C je A
 if (a<=0.0f) return(A);

 b=v2.magnitude(); // dlzka vektora A->B

 // bod C lezi za koncom ciary => najblizsim bodom k C je B
 if (a>=b) return(B);

 // vypocet polohy bodu na ciare A-B
 return(A+v1*a);
}
