#ifndef _MATH3D_H_
#define _MATH3D_H_

// rozne geometricke a matematicke funkcie *********************************

// minimum a maximum 2 hodnot
#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif
template <class T>
static inline T min(T a,T b) { return(a < b ? a : b); }
template <class T>
static inline T max(T a,T b) { return(a > b ? a : b); }

#ifdef swap
#undef swap
#endif
// vymena hodnot 2 premennych
template <class T>
static inline void swap(T &a,T &b) { T t=a; a=b; b=t; }

// konstanty
#define PI    (3.141592653589f)
#define PI180 (0.017453292519f)
#define PI2   (1.570796326794f)
// prevod stupne<->radiany
#define DEG2RAD(a) (PI/180.0f*(a))
#define RAD2DEG(a) (180.0f/PI*(a))

// 2-rozmerny vektor *******************************************************
class vec2 {
 public:

  // udaje -----------------------------------------------------------------
  union {
   struct {
    float x,y;
   };
   float v[2];
  };

  // konstruktory ----------------------------------------------------------
  vec2(float A=0.0f,float B=0.0f):x(A),y(B) {}
  vec2(const vec2 &V):x(V.x),y(V.y) {}
  vec2(float *V):x(V[0]),y(V[1]) {}
  vec2(int V[2]):x(V[0]),y(V[1]) {}
};

// 3-rozmerny vektor *******************************************************
class vec3 {
 public:

  // udaje -----------------------------------------------------------------
  union {
   struct {
    float x,y,z;
   };
   float v[3];
  };

  // konstruktory ----------------------------------------------------------
  vec3(float A=0.0f,float B=0.0f,float C=0.0f):x(A),y(B),z(C) {}
  vec3(const vec3 &V):x(V.x),y(V.y),z(V.z) {}
  vec3(float *V):x(V[0]),y(V[1]),z(V[2]) {}
  vec3(int V[3]):x(V[0]),y(V[1]),z(V[2]) {}
  vec3(const vec3 &A,const vec3 &B):x(B.x-A.x),y(B.y-A.y),z(B.z-A.z) {}
  explicit vec3(int A):x(A),y(A),z(A) {}
  explicit vec3(float A):x(A),y(A),z(A) {}

  // operatory -------------------------------------------------------------
  // index
  float &operator[](int I)       { return v[I]; }
  float  operator[](int I) const { return v[I]; }
  // priradenie hodnot prvkom vektora
  const vec3 &operator=(const vec3 &V) { x=V.x; y=V.y; z=V.z; return(*this); }
  void set(float A,float B,float C) { x=A; y=B; z=C; }
  vec3 &set(int I,float F) { v[I]=F; return(*this); }
  // rovnost vektorov
  const bool operator==(const vec3 &V) const { return((x==V.x) && (y==V.y) && (z==V.z)); }
  // nerovnost vektorov
  const bool operator!=(const vec3 &V) const { return(!(*this==V)); }
  // scitanie vektorov
  const vec3 operator+(const vec3 &V) const { return(vec3(x+V.x,y+V.y,z+V.z)); }
  vec3 &add(const vec3 &V) { x+=V.x; y+=V.y; z+=V.z; return(*this); }
  // pricitanie skalara k prvkom vektora
  vec3 &add(float S) { x+=S; y+=S; z+=S; return(*this); }
  // pricitanie k vektoru
  const vec3 &operator+=(const vec3 &V) { x+=V.x; y+=V.y; z+=V.z; return(*this); }
  // odcitanie vektorov
  const vec3 operator-(const vec3 &V) const { return(vec3(x-V.x,y-V.y,z-V.z)); }
  vec3 &sub(const vec3 &V) { x-=V.x; y-=V.y; z-=V.z; return(*this); }
  // odcitanie skalara od prvkov vektora
  vec3 &sub(float S) { x-=S; y-=S; z-=S; return(*this); }
  // odcitanie od vektora
  const vec3 &operator-=(const vec3 &V) { x-=V.x; y-=V.y; z-=V.z; return(*this); }
  // nasobenie skalarom
  const vec3 &operator*=(const float S) { x*=S; y*=S; z*=S; return(*this); }
  const vec3 operator*(const float S) const { return(vec3(x*S,y*S,z*S)); }
  vec3 &mul(float F) { x*=F; y*=F; z*=F; return(*this); }
  // delenie skalarom
  const vec3 &operator/=(const float S) { const float recip=1.0f/S; x*=recip; y*=recip; z*=recip; return(*this); }
  const vec3 operator/(const float S) const { const float recip=1.0f/S; return(vec3(x*recip,y*recip,z*recip)); }
  vec3 &div(float F) { if (F==0.0f) F=1.0f; x/=F; y/=F; z/=F; return(*this); }
  // nasobenie vektorov
  const vec3 operator*(const vec3 &V) const { return(vec3(x*V.x,y*V.y,z*V.z)); }
  // kladne znamienko vektora
  const vec3 operator+() const { return(vec3(*this)); }
  // zaporne znamienko vektora
  const vec3 operator-() const { return(vec3(-x,-y,-z)); }
  vec3 &neg() { return(mul(-1.0f)); }

  // funkcie ---------------------------------------------------------------
  // vektorovy sucin
  const vec3 cross(const vec3 &V) const { return(vec3(y*V.z-z*V.y,z*V.x-x*V.z,x*V.y-y*V.x)); }
  vec3 &cross(const vec3 &V1,const vec3 &V2) {
   x=V1.y*V2.z-V1.z*V2.y; y=V1.z*V2.x-V1.x*V2.z; z=V1.x*V2.y-V1.y*V2.x;
   return(*this);
  }
  // skalarny sucin
  float dot(const vec3 &V) const { return(x*V.x+y*V.y+z*V.z); }
  // (dlzka vektora)^2
  float squaredlen() const { return(x*x+y*y+z*z); }
  // dlzka vektora
  float magnitude() const { return(sqrtf(squaredlen())); }
  // jednotkovy vektor
  const vec3 unit() const { return((*this)/magnitude()); }
  // normalizacia vektora
  vec3 &normalize() { div(magnitude()); return(*this); }
  // uhol medzi vektormi
  inline const float angle(const vec3 &V) const { return(acosf(dot(V))); }
  // zrkadlova projekcia 3D bodu do roviny definovanej normalou
  // rovina prechadza zaciatkom suradnicovej sustavy
  const vec3 reflect(const vec3 &N) const { return(*this-N*2.0f*dot(N)); }
  vec3 &reflect(const vec3 &N) { float k=2.0f*dot(N); x-=N.x*k; y-=N.y*k; z-=N.z*k; return(*this); }
  // rotacia vektora okolo normaly
  const vec3 rotate(const float A,const vec3 &N) const {
   const float cosine=cosf(A); const float sine=sinf(A);
   return(vec3((*this)*cosine+((N*(*this))*(1.0f-cosine))*N+(cross(N))*sine));
  }
  // nulovy vektor?
  bool iszero() const { return(x==0 && y==0 && z==0); }
  // normalizovany vektor?
  bool isnormalized() const { float m=squaredlen(); return(m>0.99f && m<1.01f); }
  // vzdialenost dvoch 3D bodov
  float dist(const vec3 &V) const { vec3 t; return(dist(V,t)); }
  float dist(const vec3 &V1,vec3 &V2) const { V2=(*this); V2.sub(V1); return(V2.magnitude()); }
  // reject
  bool reject(const vec3 &V,float MAX) { return(x>V.x+MAX || x<V.x-MAX || y>V.y+MAX || y<V.y-MAX); }
  // rescale
  void rescale(float K) { float mag=magnitude(); if (mag>1e-6f) mul(K/mag); }
  // projekcia 3D bodu do roviny
  // rovina prechadza zaciatkom suradnicovej sustavy
  vec3 &project(const vec3 &N) { float k=dot(N); x-=k*N.x; y-=k*N.y; z-=k*N.z; return(*this); }
  vec3 &projectxydir(const vec3 &N) { if (N.z) z=-(x*N.x/N.z+y*N.y/N.z); return(*this); }
  vec3 &projectxy(const vec3 &N) {
   float m=squaredlen(),k=dot(N);
   projectxydir(N);
   rescale(sqrtf(max(m-k*k,0.0f)));
   return(*this);
  }
  vec3 &projectxy(const vec3 &N,float THRESHOLD) {
   float m=squaredlen(),k=min(dot(N),THRESHOLD);
   projectxydir(N);
   rescale(sqrtf(max(m-k*k,0.0f)));
   return(*this);
  }
  // rotacia okolo osi X Y Z
  void rotate_around_x(float A) { *this=vec3(x,                  cosf(A)*y-sinf(A)*z,cosf(A)*z+sinf(A)*y); }
  void rotate_around_y(float A) { *this=vec3(cosf(A)*x-sinf(A)*z,y,                  cosf(A)*z+sinf(A)*x); }
  void rotate_around_z(float A) { *this=vec3(cosf(A)*x-sinf(A)*y,cosf(A)*y+sinf(A)*x,z); }
  // rotacia okolo osi O
  void rotate(float A,const vec3 &O) {
   float c=cosf(A),s=sinf(A);
   rotate(c,s,O);
  }
  void rotate(float C,float S,const vec3 &O) {
   *this=vec3(x*(O.x*O.x*(1-C)+C)    +y*(O.x*O.y*(1-C)-O.z*S)+z*(O.x*O.z*(1-C)+O.y*S),
              x*(O.y*O.x*(1-C)+O.z*S)+y*(O.y*O.y*(1-C)+C)    +z*(O.y*O.z*(1-C)-O.x*S),
              x*(O.x*O.z*(1-C)-O.y*S)+y*(O.y*O.z*(1-C)+O.x*S)+z*(O.z*O.z*(1-C)+C));
  }
};

// 4-rozmerny vektor *******************************************************
class vec4 {
 public:

  // udaje -----------------------------------------------------------------
  union {
   struct {
    float x,y,z,w;
   };
   float v[4];
  };

  // konstruktory ----------------------------------------------------------
  vec4(float A=0.0f,float B=0.0f,float C=0.0f,float D=0.0f):x(A),y(B),z(C),w(D) {}
  vec4(const vec4 &V):x(V.x),y(V.y),z(V.z),w(V.w) {}
  vec4(float *V):x(V[0]),y(V[1]),z(V[2]),w(V[3]) {}
  vec4(int V[4]):x(V[0]),y(V[1]),z(V[2]),w(V[3]) {}

  // operatory -------------------------------------------------------------
  // index
  float &operator[](int I)       { return v[I]; }
  float  operator[](int I) const { return v[I]; }
  // priradenie hodnot prvkom vektora
  const vec4 &operator=(const vec4 &V) { x=V.x; y=V.y; z=V.z; w=V.w; return(*this); }
  void set(float A,float B,float C,float D) { x=A; y=B; z=C; w=D; }
};

// 16-prvkova matica *******************************************************
class mat16 {
 public:

  // udaje -----------------------------------------------------------------
  float m[16];

  // konstruktory ----------------------------------------------------------
  mat16() {
   m[0]=1.0f; m[4]=0.0f; m[8] =0.0f; m[12]=0.0f;
   m[1]=0.0f; m[5]=1.0f; m[9] =0.0f; m[13]=0.0f;
   m[2]=0.0f; m[6]=0.0f; m[10]=1.0f; m[14]=0.0f;
   m[3]=0.0f; m[7]=0.0f; m[11]=0.0f; m[15]=1.0f;
  }
  mat16(const mat16 &M) {
   (*this)=M;
  }

  // operatory -------------------------------------------------------------
  // index
  float &operator[](const int I) { return m[I]; }
  // nasobenie matic, vysledok sa ulozi do povodnej matice
  const mat16 operator*=(const mat16 M) {
   m[0] =m[0]*M.m[0] +m[4]*M.m[1] +m[8] *M.m[2];
   m[1] =m[1]*M.m[0] +m[5]*M.m[1] +m[9] *M.m[2];
   m[2] =m[2]*M.m[0] +m[6]*M.m[1] +m[10]*M.m[2];
   m[3] =m[3]*M.m[0] +m[7]*M.m[1] +m[11]*M.m[2];

   m[4] =m[0]*M.m[4] +m[4]*M.m[5] +m[8] *M.m[6];
   m[5] =m[1]*M.m[4] +m[5]*M.m[5] +m[9] *M.m[6];
   m[6] =m[2]*M.m[4] +m[6]*M.m[5] +m[10]*M.m[6];
   m[7] =m[3]*M.m[4] +m[7]*M.m[5] +m[11]*M.m[6];

   m[8] =m[0]*M.m[8] +m[4]*M.m[9] +m[8] *M.m[10];
   m[9] =m[1]*M.m[8] +m[5]*M.m[9] +m[9] *M.m[10];
   m[10]=m[2]*M.m[8] +m[6]*M.m[9] +m[10]*M.m[10];
   m[11]=m[3]*M.m[8] +m[7]*M.m[9] +m[11]*M.m[10];

   m[12]=m[0]*M.m[12]+m[4]*M.m[13]+m[8] *M.m[14];
   m[13]=m[1]*M.m[12]+m[5]*M.m[13]+m[9] *M.m[14];
   m[14]=m[2]*M.m[12]+m[6]*M.m[13]+m[10]*M.m[14];
   m[15]=m[3]*M.m[12]+m[7]*M.m[13]+m[11]*M.m[14];

   return(*this);
  }
  // nasobenie matic, vysledok sa ulozi do novej matice
  const mat16 operator*(const mat16 M) {
   mat16 mat;

   mat.m[0] =m[0]*M.m[0] +m[4]*M.m[1] +m[8] *M.m[2];
   mat.m[1] =m[1]*M.m[0] +m[5]*M.m[1] +m[9] *M.m[2];
   mat.m[2] =m[2]*M.m[0] +m[6]*M.m[1] +m[10]*M.m[2];
   mat.m[3] =m[3]*M.m[0] +m[7]*M.m[1] +m[11]*M.m[2];

   mat.m[4] =m[0]*M.m[4] +m[4]*M.m[5] +m[8] *M.m[6];
   mat.m[5] =m[1]*M.m[4] +m[5]*M.m[5] +m[9] *M.m[6];
   mat.m[6] =m[2]*M.m[4] +m[6]*M.m[5] +m[10]*M.m[6];
   mat.m[7] =m[3]*M.m[4] +m[7]*M.m[5] +m[11]*M.m[6];

   mat.m[8] =m[0]*M.m[8] +m[4]*M.m[9] +m[8] *M.m[10];
   mat.m[9] =m[1]*M.m[8] +m[5]*M.m[9] +m[9] *M.m[10];
   mat.m[10]=m[2]*M.m[8] +m[6]*M.m[9] +m[10]*M.m[10];
   mat.m[11]=m[3]*M.m[8] +m[7]*M.m[9] +m[11]*M.m[10];

   mat.m[12]=m[0]*M.m[12]+m[4]*M.m[13]+m[8] *M.m[14]+m[12];
   mat.m[13]=m[1]*M.m[12]+m[5]*M.m[13]+m[9] *M.m[14]+m[13];
   mat.m[14]=m[2]*M.m[12]+m[6]*M.m[13]+m[10]*M.m[14]+m[14];
   mat.m[15]=m[3]*M.m[12]+m[7]*M.m[13]+m[11]*M.m[14]+m[15];

   return(mat);
  }
  // nasobenie vektora a matice
  const vec3 operator*(const vec3 V) {
   float a,b,c;
   a=V.x*m[0]+V.y*m[4]+V.z*m[8] +m[12];
   b=V.x*m[1]+V.y*m[5]+V.z*m[9] +m[13];
   c=V.x*m[2]+V.y*m[6]+V.z*m[10]+m[14];
   return(vec3(a,b,c));
  }

  // funkcie ---------------------------------------------------------------
  // priradenie prvkov matici
  void set(const float M[16]) {
   m[0]=M[0]; m[4]=M[4]; m[8] =M[8];  m[12]=M[12];
   m[1]=M[1]; m[5]=M[5]; m[9] =M[9];  m[13]=M[13];
   m[2]=M[2]; m[6]=M[6]; m[10]=M[10]; m[14]=M[14];
   m[3]=M[3]; m[7]=M[7]; m[11]=M[11]; m[15]=M[15];
  }
  // ziskanie prvkov matice
  float *get() {
   return(m);
  }
  // jednotkova matica
  void identity() {
   m[0]=1.0f; m[4]=0.0f; m[8] =0.0f; m[12]=0.0f;
   m[1]=0.0f; m[5]=1.0f; m[9] =0.0f; m[13]=0.0f;
   m[2]=0.0f; m[6]=0.0f; m[10]=1.0f; m[14]=0.0f;
   m[3]=0.0f; m[7]=0.0f; m[11]=0.0f; m[15]=1.0f;
  }
  // posun
  void translate(float X,float Y,float Z) {
   m[12]+=X; m[13]+=Y; m[14]+=Z;
  }
  // rotacia
  void rotate(float X,float Y,float Z) {
   float cx,cy,cz,sx,sy,sz;
   mat16 mat;

   cx=cosf(X); sx=sinf(X);
   cy=cosf(Y); sy=sinf(Y);
   cz=cosf(Z); sz=sinf(Z);

   mat.m[0] =cy*cz;
   mat.m[1] =cy*sz;
   mat.m[2] =-sy;
   mat.m[4] =(sx*sy*cz)-(cx*sz);
   mat.m[5] =(sx*sy*sz)+(cx*cz);
   mat.m[6] =sx*cy;
   mat.m[8] =(cx*sy*cz)+(sx*sz);
   mat.m[9] =(cx*sy*sz)-(sx*cz);
   mat.m[10]=cx*cy;

   (*this)=(*this)*mat;
  }
};

// trieda premennych a funkcii (objektov) pre pracu s matematikou
class CMath3d {
public:
 CMath3d();  // konstruktor
 ~CMath3d(); // destruktor

 // kolizia svetelneho luca a trojuholnika
 bool ray_tri_collision(vec3 P,vec3 L,vec3 N,float D,vec4 CN0,vec4 CN1,vec4 CN2);
 // test polohy bodu a trojuholnika
 bool point_in_triangle(vec3 P,vec4 CN0,vec4 CN1,vec4 CN2);
 // kolizia bodu so stranami trojuholnika
 bool edge_collision(vec3 *P,vec3 *S,vec3 A,vec3 B,vec3 C,float R);
 // vypocet polohy bodu na ciare A-B, ktory je najblizsie k bodu C
 vec3 point_on_line(vec3 A,vec3 B,vec3 C);
private:
 vec3 v0,v1,v2; // pomocne vektory

 float a,b; // pomocne premenne
};

#endif
